/* File:        cli.c
 * Version:     002
 * Date:        Jul 05, 2009
 * Author:      hueb
 * Description: Command line interpreter implementation.
 *
 * Change history:
 * 001 Jun 12, 2009 hueb Original issue
 * 002 Jul 05, 2009 hueb Extensions for UART cli introduction, formal
 *                       corrections
 */

/******************************************************************************/

/* Includes */

#include "system.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/******************************************************************************/

/* Defintions */

#define PRINTABLE(c) ((31 < c && c < 127)) ? c : '.'
#define VGA 0

/******************************************************************************/

/* Globals */
CliInterface vga_cli;
CliInterface com1_cli;


/******************************************************************************/

/* Funcions */

/* Function:    cli_puts
 * Description: Prints a string either on VGA or on UART. */
void cli_puts(CliInterface * cli, char * s)
{
	if (VGA == cli->interface_info)
		vga_puts(s);
	else
		uart_puts(cli->interface_info, s);
}

/* Function:    cli_reset
 * Description: Performs a reset on a CLI. */
void cli_reset(CliInterface * cli)
{
	n_memset(cli->cli_cmd_buf, 0, sizeof(cli->cli_cmd_buf));
	cli->cli_cmd_ptr = cli->cli_cmd_buf;
}

/* Function:    cli_usage
 * Description: Shows help information in cli. */
void cli_usage(CliInterface * cli, CliCommand cmd)
{
	char s[128];

	switch (cmd)
	{
		case CliCmdDisplay:
			sprintf(s, "dis  - Display memory content. Syntax: dis 0x12345678\r\n");
			break;
		case CliCmdHelp:
			sprintf(s, "help - Display help. Syntax: help [command]\r\n");
			break;
		case CliCmdReboot:
			sprintf(s, "rbt  - Reboot the system.\r\n");
			break;
		default:
			sprintf(s, "cli command unkown.\r\n");
	}

	cli_puts(cli, s);
}

/* Function:    scan_clibuf_on_command
 * Description: Processes a string from the command line and executes the
 *              command. */
void scan_clibuf_on_command(CliInterface * cli)
{
	unsigned char * cmdstr = cli->cli_cmd_buf;

	/* Display command */
	if ('d' == cmdstr[0] && 'i' == cmdstr[1] && 's' == cmdstr[2] && ' ' == cmdstr[3])
	{
		char * token = NULL;
		int i = 0;
		int ii = 0;
		int l = 0;
		unsigned char * p = NULL;

		token = strtok(&cmdstr[4], " ");

		/* Empty token. */
		if (NULL == token)
		{
			cli_puts(cli, "Syntax error 1.\r\n");
			cli_usage(cli, CliCmdDisplay);
			return;
		}

		/* Does token start with "0x"? */
		if ('0' == token[0] && ('x' == token[1] || 'X' == token[1]))
		{
			token += 2;
		}
		else
		{
			cli_puts(cli, "Syntax error 2.\r\n");
			cli_usage(cli, CliCmdDisplay);
			return;
		}

		/* Has token a length of 8? */
		l = strlen(token) - 1;
		if (8 != l)
		{
			cli_puts(cli, "Syntax error 3.\r\n");
			cli_usage(cli, CliCmdDisplay);
			return;
		}

		/* Does token only consist of hexadecimal signs? */
		for (i=0; i<l; i++)
		{
			char hexSigns[] = {'0', '1', '2', '3', 
		                       '4', '5', '6', '7', 
                		       '8', '9', 'A', 'B', 
                       			'C', 'D', 'E', 'F', 
					'a', 'b', 'c', 'd', 
					'e', 'f'};

			/* Compare with each hexSign */
			for (ii=0; ii<sizeof(hexSigns); ii++)
				if (token[i] == hexSigns[ii])
					break;

			/* If end of array is reached and current sign is not
			   within hexSigns, sign is not a hexedecimal sign. */
			if (token[i] != hexSigns[ii])
			{
				cli_puts(cli, "Syntax error 4.\r\n");
				cli_usage(cli, CliCmdDisplay);
				return;
			}

			/* Convert lower case into upper case. */
			if ('a' <= token[i])
				token[i] -= 32;
		}

		/* Convert string into pointer. */
		for (i=0; i<l; i++)
		{
			unsigned char c = (unsigned char) token[i];
			if ('9' >= c)
				p += (c - 48) << ((l-1)-i) * 4;
			else
				p += (c - 55) << ((l-1)-i) * 4;
			/* 0xABCDEF01 */
		}

		printmem(cli, p, 16);

		return;
	}

	/* Reboot command */
	if ('r' == cmdstr[0] && 'b' == cmdstr[1] && 't' == cmdstr[2])
	{
		cli_puts(cli, "Reboot ...\r\n");
		cli_reboot();
		return;
	}

	/* Help command */
	if ('h' == cmdstr[0] && 'e' == cmdstr[1] && 'l' == cmdstr[2] && 'p' == cmdstr[3])
	{
		char * token = NULL;

		token = strtok(&cmdstr[4], " ");

		if ('d' == token[0] && 'i' == token[1] && 's' == token[2])
			cli_usage(cli, CliCmdDisplay);
		else if ('r' == token[0] && 'b' == token[1] && 't' == token[2])
			cli_usage(cli, CliCmdReboot);
		else
		{
			cli_puts(cli, "NAOS cli help\r\n\r\n");
			cli_usage(cli, CliCmdDisplay);
			cli_usage(cli, CliCmdHelp);
			cli_usage(cli, CliCmdReboot);
		}
		
		return;
	}

	if ('\n' != cmdstr[0])
		cli_puts(cli, "Syntax error. Type help [command] to get further information.\r\n");
}

/* Function:    cli_putc
 * Description: Puts a character from user's input on the screen, depending on
 *              it's printing character. Not the whole keyboard layout is
 *              implemented yet. */
void cli_putc(CliInterface * cli, unsigned char c)
{
	/* Backspace */
	if ('\b' == c)
	{
		/* Don't go back on col 0. */
		if (cli->cli_cmd_ptr <= cli->cli_cmd_buf)
			return;

		/* Decrease cli command pointer and erase char on screen. */
		cli->cli_cmd_ptr--;
		if (VGA == cli->interface_info)
		{
			vga_putc(' ');
			vga_putc(c);
		}
		else
		{
			uart_putc(cli->interface_info, ' ');
			uart_putc(cli->interface_info, c);
		}
		return;
	}

	/* If it is printable or ENTER, add char to cli command. */
	if (ISPRINTABLE(c))
		*cli->cli_cmd_ptr++ = c;
	else
		return;
	
	/* If cli command buffer is full, reset. */
	if (sizeof(cli->cli_cmd_buf) <= (size_t)(cli->cli_cmd_ptr - cli->cli_cmd_buf))
	{
		cli_reset(cli);
		cli_puts(cli, "\r\n *** cli: Command exceeded cli buffer. Reset.\r\n");
	}

	/* If ENTER was hit, terminate string, scan cli buffer on commands and reset. */
	if ('\n' == c)
	{
		*cli->cli_cmd_ptr = 0;
		scan_clibuf_on_command(cli);
		cli_reset(cli);
	}
}

/* Function:    printmem
 * Description: Prints a memory area's content on the sreen. */
void printmem(CliInterface * cli, unsigned char * p, unsigned long len)
{
	int i = 0;
	char buf[96];
	
	for (i=0; i<len/16; i++)
	{
		memset(buf, 0, sizeof(buf));
		sprintf(buf, "0x%08X %02X %02X %02X %02X  %02X %02X %02X %02X  %02X %02X %02X %02X  %02X %02X %02X %02X  %c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c\r\n", 
			(unsigned)(p+i*16), 
			p[i+0], p[i+1], p[i+2], p[i+3], 
			p[i+4], p[i+5], p[i+6], p[i+7], 
			p[i+8], p[i+9], p[i+10], p[i+11], 
			p[i+12], p[i+13], p[i+14], p[i+15], 
			PRINTABLE(*p), PRINTABLE(p[i+1]), PRINTABLE(p[i+2]), PRINTABLE(p[i+3]), 
			PRINTABLE(p[i+4]), PRINTABLE(p[i+5]), PRINTABLE(p[i+6]), PRINTABLE(p[i+7]), 
			PRINTABLE(p[i+8]), PRINTABLE(p[i+9]), PRINTABLE(p[i+10]), PRINTABLE(p[i+11]), 
			PRINTABLE(p[i+12]), PRINTABLE(p[i+13]), PRINTABLE(p[i+14]), PRINTABLE(p[i+15]));
		cli_puts(cli, buf);
	}
}

/* GNU reset mechanisms.
 * Taken from GNU / Linux kernel 2.6.17:
 * Linux-2.6.17/arch/i386/kernel/reboot.c
 * Incredible shell code and inline assembler shit. Thanks! */

static unsigned long long real_mode_gdt_entries [3] = 
{
	0x0000000000000000ULL, /* NULL descriptor. */
	0x00009A000000FFFFULL, /* 16 bit real mode 64k code at 0x00000000. */
	0x000092000100FFFFULL  /* 16 bit real mode 64k data at 0x00000100. */
};

/* Nameless type for GDT and IDT. */
struct
{
	unsigned short 		size __attribute__ ((packed));
	unsigned long long * 	base __attribute__ ((packed));
}

/* Global GDT and IDT for code in start.asm */
real_mode_gdt = {sizeof (real_mode_gdt_entries) - 1, real_mode_gdt_entries}, 
real_mode_idt = {0x03FF, NULL}, 
no_idt = {0, NULL};

/* 16 bit protected mode code.
 * Disables paging.
 * Disables the cache.
 * Switches to real mode.
 * Jumps to the BIOS reset code.
 * 
 * The instruction that switches to real mode by writing to CR0 must be 
 * followed immediatly by by a far jump instruction. This sets CS to a valid 
 * value for real mode and flushes the prefetch queue to avoid running 
 * instructions that have already been decoded in protected mode.
 *
 * Clears all the flags except ET, espcially PG (paging), PE (protected mode 
 * enable) ant TS (task switch for coprocessor state save). Flushees the TLB 
 * after paging has been disabled. Sets CD and NW to disable the cache on a 
 * 486 and invalidates the cache. This is more like the state of a 486 after 
 * reset. */
static unsigned char real_mode_switch[] = 
{
	0x66, 0x0F, 0x20, 0xC0, 			/* movl %cr0,%eax 		*/
	0x66, 0x83, 0xE0, 0x11, 			/* andl $0x00000011,%eax	*/
	0x66, 0x0D, 0x00, 0x00, 0x00, 0x60, 		/* orl  $0x60000000,%eax	*/
	0x66, 0x0F, 0x22, 0xC0, 			/* movl %eax,%cr0		*/
	0x66, 0x0F, 0x22, 0xD8, 			/* movl %eax,%cr3		*/
	0x66, 0x0F, 0x20, 0xC3, 			/* movl %cr0,%ebx		*/
	0x66, 0x81, 0xE3, 0x00, 0x00, 0x00, 0x60, 	/* andl $0x60000000		*/
	0x74, 0x02, 					/* jz   f			*/
	0x0F, 0x09, 					/* wbinvd			*/
	0x24, 0x10, 					/* f: andb $0x10,al		*/
	0x66, 0x0F, 0x22, 0xC0				/* movl %eax,%cr0		*/
};

static unsigned char jump_to_bios [] = 
{
	0xEA, 0x00, 0x00, 0xFF, 0xFF	/* ljmp $0xFFFF,$0x0000 */
};

extern void gnu_idt_load();

/* Switch to real mode and then execute the jump_to_bios() code. */
void machine_real_restart()
{
	unsigned char * code = jump_to_bios;
	int length = sizeof (jump_to_bios);

	/* GNU / Linux does more here ... */

	asm volatile ("cli");

	/* Write BIOS code for warm reset (0x1234) to memory at address
	 * 0x0472. */
	*((unsigned short *) 0x0472) = 0x1234;

	/* Copy code into the lower 64k memory. */
	memcpy ((void *) (0x1000 - sizeof (real_mode_switch) - 100), 
		real_mode_switch, 
		sizeof (real_mode_switch));
	memcpy ((void *) (0x1000 - 100), 
		code, 
		length);

	/* Set up IDT and GDT for real mode. */
	gnu_idt_load();
	gnu_gdt_flush();

	/* Prepare data segment registers and descriptors for real mode. */
	asm volatile ("movl $0x0010,%%eax\n"
			"\tmovl %%eax,%%ds\n"
			"\tmovl %%eax,%%es\n"
			"\tmovl %%eax,%%fs\n"
			"\tmovl %%eax,%%gs\n"
			"\tmovl %%eax,%%ss" : : : "eax");

	/* Jump to the copied code. */
	asm volatile ("ljmp $0x0008,%0" : : "i" ((void *) (0x1000 - sizeof (real_mode_switch) - 100)));
}

void gnu_reset()
{
	machine_real_restart();
}

/* End of GNU reset mechanisms. */

/* Function:    cli_reboot
 * Description: CLI reboot command implementation. */
void cli_reboot()
{
	/* Version 1: pulse RESET line from 8042 controller.
	   After that, IP > 0x0000FFFF -> Error and HLT. :-( */
	/* unsigned char good = 0x02;
	while ((good & 0x02) != 0)
		good = inportb(0x64);
	outportb(0x60, 0xFE);
	__asm__ __volatile__ ("hlt"); */ /* if that didn't work, halt the CPU */

	/* Version 2: */
	/* Reboot by stwitching back from protected mode to real mode and jump
	 * to Address 0xFFFF0. It contains the jmp sequence to the BIOS startup
	 + code.
	   Code located in start.asm. -> Bochs exception :-( */
	/* reboot(); */

	/* Version 3: Produce a triple fault by setting an IDT of size 0.
	   This leads to a reset. Not the nice way, but it work's. */
	/* idt_reset(); */

	/* Version 4: The GNU way */
	gnu_reset();
}

void cli_init()
{
	/* VGA CLI */
	n_memset((unsigned char *) &vga_cli, 0, sizeof(vga_cli));
	vga_cli.cli_cmd_ptr = vga_cli.cli_cmd_buf;
	vga_cli.interface_info = VGA;

	/* UART / COM1 CLI */
	n_memset((unsigned char *) &com1_cli, 0, sizeof(com1_cli));
	com1_cli.cli_cmd_ptr = com1_cli.cli_cmd_buf;
	com1_cli.interface_info = UART_PORT_COM1;
}
