/* File:        idt.c
 * Version:     002
 * Date:        May 11, 2009
 * Author:      hueb
 * Description: Implementation of the interrupt descriptor table.
 *
 * Change history:
 * 001 Jan 09, 2009 hueb Original issue
 * 002 May 11, 2009 hueb Adding reboot code.
 */

/******************************************************************************/

/* Includes */

#include "system.h"

/******************************************************************************/

/* Globals */

/* Defines an IDT entry. */
struct idt_entry
{
    unsigned short base_lo;
    unsigned short sel;        /* Our kernel segment goes here! */
    unsigned char always0;     /* This will ALWAYS be set to 0! */
    unsigned char flags;       /* Set using the above table! */
    unsigned short base_hi;
} __attribute__((packed));

struct idt_ptr
{
    unsigned short limit;
    unsigned int base;
} __attribute__((packed));

/* Declare an IDT of 256 entries. Although we will only use the
 * first 32 entries in this tutorial, the rest exists as a bit
 * of a trap. If any undefined IDT entry is hit, it normally
 * will cause an "Unhandled Interrupt" exception. Any descriptor
 * for which the 'presence' bit is cleared (0) will generate an
 * "Unhandled Interrupt" exception. */
struct idt_entry idt[256];
struct idt_ptr idtp;

/******************************************************************************/

/* Prototypes */

/* This exists in 'start.asm', and is used to load our IDT. */
extern void idt_load();

/******************************************************************************/

/* Functions */

/* Function:    idt_set_gate
 * Description: Use this function to set an entry in the IDT. Alot simpler
 *              than twiddling with the GDT. ;) */
void idt_set_gate(unsigned char num, unsigned long base, unsigned short sel, unsigned char flags)
{
    /* We'll leave you to try and code this function: take the
    *  argument 'base' and split it up into a high and low 16-bits,
    *  storing them in idt[num].base_hi and base_lo. The rest of the
    *  fields that you must set in idt[num] are fairly self-
    *  explanatory when it comes to setup */
	idt[num].base_lo = (unsigned short) (base & 0x0000FFFF);
	idt[num].base_hi = (unsigned short) ((base >> 16) & 0x0000FFFF);
	idt[num].sel = sel;
	idt[num].always0 = 0;
	idt[num].flags = flags;
}

/* Function:    idt_set_gate
 * Description: Installs the IDT. */
void idt_install()
{
    /* Sets the special IDT pointer up, just like in 'gdt.c' */
    idtp.limit = (sizeof (struct idt_entry) * 256) - 1;
    idtp.base = (unsigned int) &idt;

    /* Clear out the entire IDT, initializing it to zeros */
    n_memset((void *) &idt, 0, sizeof(struct idt_entry) * 256);

    /* Add any new ISRs to the IDT here using idt_set_gate */

    /* Points the processor's internal register to the new IDT */
    idt_load();
}

/* Function:    idt_set_gate
 * Description: Fallback reboot method: Crush the IDT. Triple fault resets the
 *              system. */
void idt_reset()
{
    idtp.limit = 0;
    idtp.base = (unsigned int) &idt;
    n_memset((void *) &idt, 0, sizeof(struct idt_entry) * 256);
    idt_load();
}
