/* File:        screen.c
 * Version:     002
 * Date:        Jul 05, 2009
 * Author:      hueb
 * Description: VGA handling.
 *
 * Change history:
 * 001 May 01, 2009 hueb Original issue
 * 002 Jul 05, 2009 hueb Formal corrections, VGA / UART CLI workaround
 */

/******************************************************************************/

/* Includes */

#include "system.h"

/******************************************************************************/

/* Globals */

/* These define our textpointer, our background and foreground
*  colors (attributes), and x and y cursor coordinates */
unsigned short *textmemptr;
int attrib = 0x0F;
int csr_x = 0, csr_y = 0;

/******************************************************************************/

/* Functions */

/* Function:    scroll
 * Description: Scrolls the screen. */
static void scroll(void)
{
    unsigned blank, temp;

    /* A blank is defined as a space... we need to give it
     * backcolor too */
    blank = 0x20 | (attrib << 8);

    /* Row 25 is the end, this means we need to scroll up */
    if(csr_y >= 25)
    {
        /* Move the current text chunk that makes up the screen
         * back in the buffer by a line */
        temp = csr_y - 25 + 1;
        n_memcpy ((void *) textmemptr, (void *) (textmemptr + temp * 80), (25 - temp) * 80 * 2);

        /* Finally, we set the chunk of memory that occupies
         * the last line of text to our 'blank' character */
        n_memsetw (textmemptr + (25 - temp) * 80, blank, 80);
        csr_y = 25 - 1;
    }
}

/* Function:    move_csr
 * Description: Updates the hardware cursor. */
static void move_csr(void)
{
    unsigned temp;

    /* The equation for finding the index in a linear
     * chunk of memory can be represented by:
     * Index = [(y * width) + x] */
    temp = csr_y * 80 + csr_x;

    /* This sends a command to indicies 14 and 15 in the
     * CRT Control Register of the VGA controller. These
     * are the high and low bytes of the index that show
     * where the hardware cursor is to be 'blinking'. To
     * learn more, you should look up some VGA specific
     * programming documents. A great start to graphics:
     * http://www.brackeen.com/home/vga */
    outportb(0x3D4, 14);
    outportb(0x3D5, temp >> 8);
    outportb(0x3D4, 15);
    outportb(0x3D5, temp);
}

/* Function:    vga_cls
 * Description: Clears the screen. */
void vga_cls()
{
    unsigned blank;
    int i;

    /* Again, we need the 'short' that will be used to
     * represent a space with color. */
    blank = 0x20 | (attrib << 8);

    /* Sets the entire screen to spaces in our current
     * color. */
    for(i = 0; i < 25; i++)
        n_memsetw (textmemptr + i * 80, blank, 80);

    /* Update out virtual cursor, and then move the
     * hardware cursor. */
    csr_x = 0;
    csr_y = 0;
    move_csr();
}

/* Function:    vga_putc
 * Description: Puts a single character on the screen. */
void vga_putc(char c)
{
    unsigned short *where;
    unsigned att = attrib << 8;

    /* Handle a backspace, by moving the cursor back one space. */
    if(c == 0x08)
    {
        if(csr_x != 0) csr_x--;
    }
    /* Handles a tab by incrementing the cursor's x, but only
     * to a point that will make it divisible by 8. */
    else if(c == 0x09)
    {
        csr_x = (csr_x + 8) & ~(8 - 1);
    }
    /* Handles a 'Carriage Return', which simply brings the
     * cursor back to the margin. */
    else if(c == '\r')
    {
        csr_x = 0;
    }
    /* We handle our newlines the way DOS and the BIOS do: we
     * treat it as if a 'CR' was also there, so we bring the
     * cursor to the margin and we increment the 'y' value. */
    else if(c == '\n')
    {
        csr_x = 0;
        csr_y++;
    }
    /* Any character greater than and including a space, is a
     * printable character. The equation for finding the index
     * in a linear chunk of memory can be represented by:
     * Index = [(y * width) + x]. */
    else if(c >= ' ')
    {
        where = textmemptr + (csr_y * 80 + csr_x);
        *where = c | att;	/* Character AND attributes: color */
        csr_x++;
    }

    /* If the cursor has reached the edge of the screen's width, we
     * insert a new line in there. */
    if(csr_x >= 80)
    {
        csr_x = 0;
        csr_y++;
    }

    /* Scroll the screen if needed, and finally move the cursor. */
    scroll();
    move_csr();
}

/* Function:    vga_puts
 * Description: Uses the above routine to output a string... */
int vga_puts(const char *text)
{
    int i = 0;

    for (i = 0; i < n_strlen(text); i++)
        vga_putc(text[i]);

    return i;
}

/* Function:    vga_puti
 * Description: Uses the above routine to output an integer... */
int vga_puti(const int value)
{
    int i = 8;
    int base = 10;
    int mod = base;
    int asciiOffset = 0x30;
    int rest = 0;

    char text[] = {0, 0, 0, 0, 
                   0, 0, 0, 0, 
                   0, 0};

/* 123456789
9 = value % 10
8 = ((value % 100) - 9) / 10
7 = ((value % 1000) - ((8*10)+9)) / 100
... */

    for (i=8; i>=0; i--)
    {
        int tmp = value % mod;
        tmp -= rest;
        rest += tmp;
        tmp /= (mod / 10);
        text[i] = tmp + asciiOffset;
        mod *= 10;
    }

    for (i = 0; i < n_strlen(text); i++)
    {
        if (0 != text[i])
            vga_putc(text[i]);
        else
            vga_putc(' ');
    }

    return i;
}

/* Function:    vga_putp
 * Description: Uses the above routine to output an pointer... */
int vga_putp(const unsigned int ptr)
{
    unsigned int i = 8;
    char text[] = {'0', 'x', 0, 0, 
                    0,   0,  0, 0, 
                    0,   0,  0};
    char hexSigns[] = {'0', '1', '2', '3', 
                       '4', '5', '6', '7', 
                       '8', '9', 'A', 'B', 
                       'C', 'D', 'E', 'F'};

/* 0x12345678
8 = value & 0x0F
7 = (value >> 4) & 0x0F
6 = (value >> 8) & 0x0F
... */

    for (i=9; i>1; i--)
    {
        unsigned int shiftMask = 0x0F;
        unsigned int tmp = (ptr >> (7-(i-2))*4) & shiftMask;
        text[i] = hexSigns[tmp];
    }

    for (i = 0; i < n_strlen(text); i++)
    {
        vga_putc(text[i]);
    }

    return i;
}

/* Function:    vga_settextcolor
 * Description: Sets the forecolor and backcolor that we will use. */
void vga_settextcolor(unsigned char forecolor, unsigned char backcolor)
{
    /* Top 4 bytes are the background, bottom 4 bytes
     * are the foreground color. */
    attrib = (backcolor << 4) | (forecolor & 0x0F);
}

/* Function:    init_video
 * Description: Sets our text-mode VGA pointer, then clears the screen
 *              for us. */
void vga_init(void)
{
    textmemptr = (unsigned short *)0xB8000;
    vga_cls();
}
