/* File:        uart.c
 * Version:     001
 * Date:        Jul 03, 2009
 * Author:      hueb
 * Description: UART implementation.
 *
 * Change history:
 * 001 Jul 03, 2009 hueb Original issue 
 */

/******************************************************************************/

/* Includes */
#include "system.h"

/******************************************************************************/

/* Globals */

extern CliInterface com1_cli;

/******************************************************************************/

/* Functions */

/* Function:    uart_received
 * Description: Returns, wether any characters are available in the UART's
 *              receive buffer. */
static int uart_received(unsigned short uart_port)
{
	return inportb(uart_port + 5) & 1;
}

/* Function:    uart_getc
 * Description: Reads a character from a UART's receive buffer. */
char uart_getc(unsigned short uart_port)
{
	while (uart_received(uart_port) == 0);
	return inportb(uart_port);
}

/* Function:    uart_is_transmit_empty
 * Description: Returns, wether a UART's transmit buffer is ready to process
 *              new data. */
static int uart_is_transmit_empty(unsigned short uart_port)
{
	return inportb(uart_port + 5) & 0x20;
}

/* Function:    uart_putc
 * Description: Writes a character to a UART's transmit buffer. */
void uart_putc(unsigned short uart_port, char c)
{
	while (uart_is_transmit_empty(uart_port) == 0);
	outportb(uart_port, c);
}

/* Function:    uart_puts
 * Description: Writes a string to a UART's transmit buffer. */
void uart_puts(unsigned short uart_port, char * s)
{
	int len = n_strlen(s);
	int i = 0;

	for (i=0; i<len; i++)
	{
		while (uart_is_transmit_empty(uart_port) == 0);
		outportb(uart_port, s[i]);
	}
}

/* Function:    uart_test
 * Description: Writes a 'Hello World!' message to COM1. */
void uart_test(unsigned short uart_port)
{
	uart_puts(UART_PORT_COM1, "Hello Naos!\r\n");
}

/* Function:    uart_detect
 * Description: Detects, which chip is used on a UART's port. Returns
 *              UART_ABSENT	on error.
 *              UART_8250	for 8259
 *              UART_16450 	for 16450 or 8250 with scratch register
 *              UART_16550 	for 16550
 *              UART_16550A 	for 16550A */
static int uart_detect(unsigned short uart_port)
{
	int x;
	int olddata;
	
	/* Check if a UART is present anyway */
	olddata = inportb(uart_port + 4);
	outportb(uart_port + 4, 0x10);
	if ((inportb(uart_port + 6) & 0xF0)) 
		return UART_ABSENT;

	outportb(uart_port + 4, 0x1F);
	if ((inportb(uart_port + 6) & 0xF0) != 0xF0)
		return UART_ABSENT;

	outportb(uart_port + 4, olddata);

	/* Next thing to do is look for the scratch register */
	olddata = inportb(uart_port + 7);
	outportb(uart_port + 7, 0x55);
	if (inportb(uart_port + 7) != 0x55)
		return UART_8250;

	outportb(uart_port + 7, 0xAA);
	if (inportb(uart_port + 7) != 0xAA)
		return UART_8250;

	/* We don't need to restore it if it's not there */
	outportb(uart_port + 7, olddata);

	/* Then check if there's a FIFO */
	outportb(uart_port + 2, 1);
	x = inportb(uart_port + 2);

	/* Some old-fashioned software relies on this! */
	outportb(uart_port + 2, 0x0);
	if ((x & 0x80) == 0)
		return UART_16450;

	if ((x & 0x40) == 0)
		return UART_16550;

	return UART_16550A;
}

/* Function:    uart_handler_com1
 * Description: Interrupthandler for COM1.*/
static void uart_handler_com1(struct regs *r)
{
	if (0 == uart_received(UART_PORT_COM1))
	{
		vga_puts("*** UART error on COM1: IRQ without incoming data.\r\n");
	}
	else
	{
		do
		{
			char c = inportb(UART_PORT_COM1);	/* Reading. */
			/* vga_puti(c); */			/* DEBUG */
			uart_putc(UART_PORT_COM1, c);		/* Mirroring. */
			cli_putc(&com1_cli, c);			/* Processing. */

			/* Simulating a COM port in BOCHS with TTY, behavior is
			 * strange: TTY sends DEL on backspace and '\r' on 
			 * ENTER. Workaround. */
			if (127 == c)
			{
				uart_putc(UART_PORT_COM1, '\b');
				cli_putc(&com1_cli, '\b');
				cli_putc(&com1_cli, '\b');
			}

			/* Completing TTY's '\r' with '\n'. */
			if ('\r' == c)
			{
				uart_putc(UART_PORT_COM1, '\n');
				cli_putc(&com1_cli, '\n');
			}
		}while (0 != uart_received(UART_PORT_COM1));
	}
}

/* Function:    uart_init
 * Description: Does initialization stuff with a COM port.
 *              Enables Received Data Available Interrupt. */
static void uart_init(unsigned short uart_port)
{
	outportb(uart_port + 1, 0x01);	/* Enable Received Data Available Interrupt. */
	outportb(uart_port + 3, 0x80);	/* Enable DLAB (set baud rate divisor). */
	outportb(uart_port + 0, 0x03);	/* Set divisor to 3 (lo byte) 38400 baud. */
	outportb(uart_port + 1, 0x00);	/*                  (hi byte). */
	outportb(uart_port + 3, 0x03);	/* 8 bits, no parity, one stop bit. */
	outportb(uart_port + 2, 0xC7);	/* Enable FIFO, clear them, with 14-byte threshold. */
	outportb(uart_port + 4, 0x0B);	/* IRQs enabled, RTS/DSR set. */
}

/* Function:    uart_install
 * Description: Detects available COM ports, detects which chip is used,
 *              carries out configuration and registers interrupt handlers.
 *              Currently implemented only for COM1. */
void uart_install()
{
	int result = UART_ABSENT;

	vga_puts("Detecting UART on COM1 ... ");
	result = uart_detect(UART_PORT_COM1);
	switch (result)
	{
		case UART_8250:		vga_puts("8250\r\n"); break;
		case UART_16450: 	vga_puts("UART_16450 / 8259 with scratch register\r\n"); break;
		case UART_16550: 	vga_puts("UART_16550\r\n"); break;
		case UART_16550A: 	vga_puts("UART_16550A\r\n"); break;
		case UART_ABSENT:
		default:
			vga_puts("failed.\r\n");
	}
	if (UART_ABSENT != result)
	{
		uart_init(UART_PORT_COM1);
		irq_install_handler(UART_IRQ_COM1, uart_handler_com1);
		uart_puts(UART_PORT_COM1, NAOS_STARTUP_STR);
		uart_puts(UART_PORT_COM1, CLI_STARTUP_STR);
	}
}
